import queue
import threading
import time
import pytest
import potassium


def test_handler():
    app = potassium.Potassium("my_app")

    @app.init
    def init():
        return {}

    @app.handler()
    def handler(context: dict, request: potassium.Request) -> potassium.Response:
        return potassium.Response(
            json={"hello": "root"},
            status=200
        )

    @app.handler("/some_path")
    def handler2(context: dict, request: potassium.Request) -> potassium.Response:
        return potassium.Response(
            json={"hello": "some_path"},
            status=200
        )

    @app.handler("/some_binary_response")
    def handler3(context: dict, request: potassium.Request) -> potassium.Response:
        return potassium.Response(
            body=b"hello",
            status=200,
            headers={"Content-Type": "application/octet-stream"}
        )

    @app.handler("/some_path_byte_stream_response")
    def handler4(context: dict, request: potassium.Request) -> potassium.Response:
        def stream():
            yield b"hello"
            yield b"world"

        return potassium.Response(
            body=stream(),
            status=200,
            headers={"Content-Type": "application/octet-stream"}
        )

    @app.handler("/some_path/child_path")
    def handler2_id(context: dict, request: potassium.Request) -> potassium.Response:
        return potassium.Response(
            json={"hello": f"some_path/child_path"},
            status=200
        )

    client = app.test_client()

    res = client.post("/", json={})
    assert res.status_code == 200
    assert res.json == {"hello": "root"}

    res = client.post("/some_path", json={})
    assert res.status_code == 200
    assert res.json == {"hello": "some_path"}

    res = client.post("/some_binary_response", json={})
    assert res.status_code == 200
    assert res.data == b"hello"
    assert res.headers["Content-Type"] == "application/octet-stream"

    res = client.post("/some_path_byte_stream_response", json={})
    assert res.status_code == 200
    assert res.data == b"helloworld"
    assert res.headers["Content-Type"] == "application/octet-stream"

    res = client.post("/some_path/child_path", json={})
    assert res.status_code == 200
    assert res.json == {"hello": "some_path/child_path"}

    res = client.post("/", data='{"key": unquoted_value}', content_type='application/json')
    assert res.status_code == 400
    # check status
    res = client.get("/__status__")
    assert res.status_code == 200
    assert res.json is not None
    assert res.json["gpu_available"] == True

# parameterized test for path collisions
@pytest.mark.parametrize("paths", [
    ("/", "",),
    ("/foo", "foo"),
])
def test_path_collision(paths):
    app = potassium.Potassium("my_app")

    @app.init
    def init():
        return {}

    @app.handler(paths[0])
    def handler(context: dict, request: potassium.Request) -> potassium.Response:
        return potassium.Response(
            json={},
            status=200
        )

    # expect exception
    with pytest.raises(potassium.RouteAlreadyInUseException):
        @app.handler(paths[1])
        def handler2(context: dict, request: potassium.Request) -> potassium.Response:
            return potassium.Response(
                json={},
                status=200
            )

def test_status():
    app = potassium.Potassium("my_app")

    resolve_background_condition = threading.Condition()

    @app.init
    def init():
        return {}

    @app.background("/background")
    def background(context: dict, request: potassium.Request):
        with resolve_background_condition:
            resolve_background_condition.wait()

    client = app.test_client()

    # send get for status
    res = client.get("/__status__", json={})

    assert res.status_code == 200
    assert res.json is not None
    assert res.json["gpu_available"] == True
    assert res.json["sequence_number"] == 0
    assert res.json["idle_time"] > 0
    assert res.json["inference_time"] == 0

    # send background post in separate thread
    res = client.post("/background", json={})
    assert res.status_code == 200

    # add a small sleep for inference time to be above 0
    time.sleep(0.1)

    # check status
    res = client.get("/__status__", json={})

    assert res.status_code == 200
    assert res.json is not None
    assert res.json["gpu_available"] == False
    assert res.json["sequence_number"] == 1
    assert res.json["idle_time"] == 0
    assert res.json["inference_time"] > 0

    # notify background thread to continue
    with resolve_background_condition:
        resolve_background_condition.notify()

    # wait for the lock to be released
    time.sleep(0.1)

    # check status
    res = client.get("/__status__", json={})

    assert res.status_code == 200
    assert res.json is not None
    assert res.json["gpu_available"] == True
    assert res.json["sequence_number"] == 1
    assert res.json["idle_time"] > 0
    assert res.json["inference_time"] == 0

    res = client.post("/this_path_does_not_exist", json={})
    assert res.status_code == 404
    res = client.get("/__status__", json={})
    assert res.status_code == 200
    assert res.json is not None
    assert res.json["gpu_available"] == True
    assert res.json["sequence_number"] == 2


def test_wait_for_background_task():
    app = potassium.Potassium("my_app")

    order_of_execution_queue = queue.Queue()
    resolve_background_condition = threading.Condition()

    @app.init
    def init():
        return {}

    @app.background("/background")
    def background(context: dict, request: potassium.Request):
        with resolve_background_condition:
            resolve_background_condition.wait()

    
    def wait_for_background_task():
        app._read_event_chan()
        order_of_execution_queue.put("background_task_completed")

    thread = threading.Thread(target=wait_for_background_task)
    thread.start()

    client = app.test_client()

    # send background post in separate thread
    order_of_execution_queue.put("send_background_task")
    res = client.post("/background", json={})
    assert res.status_code == 200

    # notify background thread to continue
    with resolve_background_condition:
        resolve_background_condition.notify()

    thread.join()

    # assert order of execution
    assert order_of_execution_queue.get() == "send_background_task"
    assert order_of_execution_queue.get() == "background_task_completed"

def test_warmup():
    app = potassium.Potassium("my_app")

    @app.init
    def init():
        return {}

    @app.handler()
    def handler(context: dict, request: potassium.Request) -> potassium.Response:
        raise Exception("should not be called")

    client = app.test_client()

    res = client.post("/_k/warmup", json={})
    assert res.status_code == 200
    assert res.json == {"warm": True}

    res = client.get("/__status__", json={})
    assert res.status_code == 200
    assert res.json is not None
    assert res.json["gpu_available"] == True
    assert res.json["sequence_number"] == 1



import os
from store import Store, RedisConfig, S3Config

# Redis store can save json serializeable objects
store = Store(
    backend="redis",
    config=RedisConfig(
        host="localhost",
        port="6379"
    )
)

objs = [
    "1",
    True,
    ["some", "list"],
    {"some": {"nested": "dict"}}
]

print("Testing json redis")
for obj in objs:
    store.set("key", obj)
    out = store.get("key")
    if out != obj:
        print("failed case", obj)

# ----

# pickle redis can save complex objects


class ComplexRedis():
    def __init__(self, a) -> None:
        self.a = a

    def __eq__(self, other):  # necessary for the equal assert later
        return self.a == other.a


objs = [
    "1",
    True,
    ["some", "list"],
    {"some": {"nested": "dict"}},
    ComplexRedis(a=1)
]

store = Store(
    backend="redis",
    config=RedisConfig(
        host="localhost",
        port="6379",
        encoding="pickle"
    )
)

print("Testing pickle redis")
for obj in objs:
    store.set("key", obj)
    out = store.get("key")
    if obj != obj:
        print("failed case", obj)


# ----


# s3
aws_access_key_id = os.environ["AWS_ACCESS_KEY_ID"]
aws_secret_access_key = os.environ["AWS_SECRET_ACCESS_KEY"]
bucket = "potassium-test"


# s3 store can save json serializeable objects
store = Store(
    backend="s3",
    config=S3Config(
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        bucket=bucket,
    )
)

objs = [
    "1",
    True,
    ["some", "list"],
    {"some": {"nested": "dict"}}
]

print("Testing json redis")
for obj in objs:
    store.set("key", obj)
    out = store.get("key")
    if out != obj:
        print("failed case", obj)

# ----

# pickle s3 can save complex objects


class ComplexS3():
    def __init__(self, a) -> None:
        self.a = a

    def __eq__(self, other):  # necessary for the equal assert later
        return self.a == other.a


objs = [
    "1",
    True,
    ["some", "list"],
    {"some": {"nested": "dict"}},
    ComplexS3(a=1)
]

store = Store(
    backend="s3",
    config=S3Config(
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        bucket=bucket,
        encoding="pickle"
    )
)

print("Testing pickle redis")
for obj in objs:
    store.set("key", obj)
    out = store.get("key")
    if obj != obj:
        print("failed case", obj)



import pytest
import potassium

def test_json_response():
    response = potassium.Response(
        status=200,
        json={"key": "value"}
    )

    assert response.status == 200
    assert response.json == {"key": "value"}
    assert response.headers["Content-Type"] == "application/json"

    response.json = {"key": "value2"}
    assert response.json == {"key": "value2"}

def test_body_response():
    response = potassium.Response(
        status=200,
        body=b"Hello, world!"
    )

    assert response.status == 200
    assert response.body == b"Hello, world!"
    assert 'Content-Type' not in response.headers

    response.json = {"key": "value2"}

    assert response.json == {"key": "value2"}
    assert response.headers["Content-Type"] == "application/json"

    





import time
from types import GeneratorType
from typing import Generator, Optional, Union
from flask import Flask, request, make_response, abort, Response as FlaskResponse
from werkzeug.serving import make_server
from threading import Thread, Lock, Condition
import functools
import traceback
import json as jsonlib
from termcolor import colored


class Endpoint():
    def __init__(self, type, func):
        self.type = type
        self.func = func


class Request():
    def __init__(self, json: dict):
        self.json = json


ResponseBody = Union[bytes, Generator[bytes, None, None]]

class Response():
    def __init__(self, status: int = 200, json: Optional[dict] = None, headers: Optional[dict] = None, body: Optional[ResponseBody] = None):
        assert json == None or body == None, "Potassium Response object cannot have both json and body set"

        self.headers = headers if headers != None else {}

        # convert json to body if not None
        if json != None:
            self.body = jsonlib.dumps(json).encode("utf-8")
            self.headers["Content-Type"] = "application/json"
        else:
            self.body = body

        self.status = status

    @property
    def json(self):
        if self.body == None:
            return None
        if type(self.body) == bytes:
            try:
                return jsonlib.loads(self.body.decode("utf-8"))
            except:
                return None
        return None
            
    @json.setter
    def json(self, json):
        self.body = jsonlib.dumps(json).encode("utf-8")
        self.headers["Content-Type"] = "application/json"


class InvalidEndpointTypeException(Exception):
    def __init__(self):
        super().__init__("Invalid endpoint type. Must be 'handler' or 'background'")


class RouteAlreadyInUseException(Exception):
    def __init__(self):
        super().__init__("Route already in use")


class Potassium():
    "Potassium is a simple, stateful, GPU-enabled, and autoscaleable web framework for deploying machine learning models."

    def __init__(self, name):
        self.name = name

        # default init function, if the user doesn't specify one
        self._init_func = lambda: {}
        # dictionary to store unlimited Endpoints, by unique route
        self._endpoints = {}  
        self._context = {}
        self._gpu_lock = Lock()
        self._background_task_cv = Condition()
        self._sequence_number = 0
        self._sequence_number_lock = Lock()
        self._idle_start_time = 0
        self._last_inference_start_time = None
        self._flask_app = self._create_flask_app()

    #
    def init(self, func):
        """init runs once on server start, and is used to initialize the app's context.
        You can use this to load models onto the GPU, set up connections, etc.
        The context is then passed into all handlers and background tasks.
        Notes:
        - this function must return a dictionary
        - the context is not persisted to disk, and will be reloaded on server restart
        - the context is not shared between multiple replicas of the app
        """

        def wrapper():
            print(colored("Running init()", 'yellow'))
            self._context = func()
            if not isinstance(self._context, dict):
                raise Exception("Potassium init() must return a dictionary")
            
        self._init_func = wrapper
        return wrapper
    
    @staticmethod
    def _standardize_route(route):
        # handle empty or None case
        if len(route) == 0 or route == None:
            route = "/"

        # prepend with "/" if not already, as router expects
        if route[0] != "/":
            route = "/" + route
        
        return route

    # handler is a blocking http POST handler
    def handler(self, route: str = "/"):
        "handler is a blocking http POST handler"

        route = self._standardize_route(route)
        if route in self._endpoints:
            raise RouteAlreadyInUseException()

        def actual_decorator(func):
            @functools.wraps(func)
            def wrapper(request):
                # send in app's stateful context if GPU, and the request
                out = func(self._context, request)

                if type(out) != Response:
                    raise Exception("Potassium Response object not returned")

                if type(out.body) != bytes and type(out.body) != GeneratorType:
                    raise Exception(
                        "Potassium Response object body must be bytes", type(out.body))

                return out

            
            self._endpoints[route] = Endpoint(type="handler", func=wrapper)
            return wrapper
        return actual_decorator

    # background is a non-blocking http POST handler
    def background(self, route: str = "/"):    
        "background is a non-blocking http POST handler"
        route = self._standardize_route(route)
        if route in self._endpoints:
            raise RouteAlreadyInUseException()

        def actual_decorator(func):
            @functools.wraps(func)
            def wrapper(request):
                # send in app's stateful context if GPU, and the request
                return func(self._context, request)
            
            self._endpoints[route] = Endpoint(
                type="background", func=wrapper)
            return wrapper
        return actual_decorator

    def test_client(self):
        "test_client returns a Flask test client for the app"
        return self._flask_app.test_client()

    # _handle_generic takes in a request and the endpoint it was routed to and handles it as expected by that endpoint
    def _handle_generic(self, endpoint, flask_request):
        # potassium rejects if lock already in use
        try:
            self._gpu_lock.acquire(blocking=False)
        except:
            res = make_response()
            res.status_code = 423
            return res

        res = None
        self._last_inference_start_time = time.time()

        try:
            req = Request(
                json=flask_request.get_json()
            )
        except:
            res = make_response()
            res.status_code = 400
            self._gpu_lock.release()
            return res

        if endpoint.type == "handler":
            try:
                out = endpoint.func(req)

                # create flask response
                res = make_response()
                res = FlaskResponse(
                    out.body, status=out.status, headers=out.headers)
            except:
                tb_str = traceback.format_exc()
                print(colored(tb_str, "red"))
                res = make_response(tb_str)
                res.status_code = 500
            self._idle_start_time = time.time()
            self._last_inference_start_time = None
            self._gpu_lock.release()
        elif endpoint.type == "background":
            # run as threaded task
            def task(endpoint, lock, req):
                try:
                    endpoint.func(req)
                except Exception as e:
                    # do any cleanup before re-raising user error
                    raise e
                finally:
                    with self._background_task_cv:
                        self._background_task_cv.notify_all()

                        self._idle_start_time = time.time()
                        self._last_inference_start_time = None
                        lock.release()

            thread = Thread(target=task, args=(endpoint, self._gpu_lock, req))
            thread.start()

            # send task start success message
            res = make_response({'started': True})
        else:
            raise InvalidEndpointTypeException()

        return res

    # WARNING: cover depends on this being called so it should not be changed
    def _read_event_chan(self) -> bool:
        """
        _read_event_chan essentially waits for a background task to finish, 
        and then returns True
        """
        with self._background_task_cv:
            # wait until the background task is done
            self._background_task_cv.wait()
        return True

    def _create_flask_app(self):
        flask_app = Flask(__name__)

        # ingest into single endpoint and spread out to multiple downstream funcs
        @flask_app.route('/', defaults={'path': ''}, methods=["POST"])
        @flask_app.route('/<path:path>', methods=["POST"])
        def handle(path):
            with self._sequence_number_lock:
                self._sequence_number += 1

            route = "/" + path
            if route not in self._endpoints:
                abort(404)

            endpoint = self._endpoints[route]
            return self._handle_generic(endpoint, request)
        
        @flask_app.route('/_k/warmup', methods=["POST"])
        def warm():
            with self._sequence_number_lock:
                self._sequence_number += 1
            res = make_response({
                "warm": True,
            })
            res.status_code = 200
            return res

        @flask_app.route('/_k/status', methods=["GET"])
        @flask_app.route('/__status__', methods=["GET"])
        def status():
            idle_time = 0
            inference_time = 0
            gpu_available = not self._gpu_lock.locked()

            if self._last_inference_start_time != None:
                inference_time = int((time.time() - self._last_inference_start_time)*1000)

            if gpu_available:
                idle_time = int((time.time() - self._idle_start_time)*1000)

            res = make_response({
                "gpu_available": gpu_available,
                "sequence_number": self._sequence_number,
                "idle_time": idle_time,
                "inference_time": inference_time,
            })

            res.status_code = 200
            return res

        return flask_app

    # serve runs the http server
    def serve(self, host="0.0.0.0", port=8000):
        print(colored("------\nStarting Potassium Server 🍌", 'yellow'))
        self._init_func()
        server = make_server(host, port, self._flask_app, threaded=True)
        print(colored(f"Serving at http://{host}:{port}\n------", 'green'))
        self._idle_start_time = time.time()
        server.serve_forever()



from typing import Optional, Union, cast
import redis
import boto3
import pickle
import json


class Entry():
    def __init__(self, value, expiration):
        self.value = value
        self.expiration = expiration

VALID_ENCODINGS = ["json", "pickle"]

class RedisConfig():
    def __init__(self, host: str, port: str, username: Optional[str] = None, password: Optional[str] = None, db: int = 0, encoding: str = "json"):
        "encoding can be 'json' or 'pickle'. JSON is default.\nPickle has better support for arbitrary python types, but using pickle with a remote redis introduces a large security risk, see https://stackoverflow.com/questions/2259270/pickle-or-json/2259351#2259351"
        # validate args
        if encoding not in VALID_ENCODINGS:
            raise ValueError(
                "redis config encoding must be one of the following:", VALID_ENCODINGS)

        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.db = db
        self.encoding = encoding


class S3Config():
    def __init__(self, aws_access_key_id, aws_secret_access_key, bucket, encoding: str = "json"):
        "encoding can be 'json' or 'pickle'. JSON is default.\nPickle has better support for arbitrary python types, but using pickle across the network to s3 introduces a large security risk, see https://stackoverflow.com/questions/2259270/pickle-or-json/2259351#2259351"
        # validate args
        if encoding not in VALID_ENCODINGS:
            raise ValueError(
                "s3 config encoding must be one of the following:", VALID_ENCODINGS)

        if aws_access_key_id is None:
            raise ValueError(
                "aws_access_key_id must be provided. It is a sensitive key, so ensure it is scoped to the bucket you want to use and not hardcoded in any open source repos.")
        if aws_secret_access_key is None:
            raise ValueError(
                "aws_secret_access_key must be provided. It is a sensitive key, so ensure it is scoped to the bucket you want to use and not hardcoded in any open source repos.")

        self.access_key = aws_access_key_id
        self.secret_access_key = aws_secret_access_key
        self.bucket = bucket
        self.encoding = encoding


class Store():
    def __init__(self, backend: str = "redis", config: Union[None, RedisConfig, S3Config] = None):
        # validate args
        backends = ["redis", "s3"]
        if backend not in backends:
            raise ValueError("backend must be one of the following:", backends)

        self.backend = backend

        if self.backend == "redis":
            if not isinstance(config, RedisConfig):
                raise ValueError("redis backends require users to bring their own redis, and configure the potassium store to use it with the config argument. For example, to use a local redis, create store with:\n\nfrom potassium.store import Store, RedisConfig\nstore = Store(backend = 'redis', config = RedisConfig(host = 'localhost', port = 6379))")
            self._redis_client = redis.Redis(
                host=config.host,
                port=int(config.port),
                username=config.username,
                password=config.password,
                db=config.db,
            )

        if self.backend == "s3":
            if not isinstance(config, S3Config):
                raise ValueError("s3 backends require users to bring their own s3 bucket, and configure the potassium store to use it with the config argument. For example, create store with:\n\nfrom potassium.store import Store, S3Config\nstore = Store(backend = 's3', config = S3Config(access_key, secret_access_key, bucket)")
            session = boto3.Session(
                aws_access_key_id=config.access_key,
                aws_secret_access_key=config.secret_access_key
            )
            self._s3_client = session.client('s3')
            self._s3_bucket = config.bucket

        assert config is not None
        self.config = config

    def get(self, key: str):
        if self.backend == "redis":
            encoded = cast(bytes, self._redis_client.get(key))
            if encoded == None:
                return None
            if self.config.encoding == "json":
                return json.loads(encoded)
            if self.config.encoding == "pickle":
                return pickle.loads(encoded)

        if self.backend == "s3":
            response = self._s3_client.get_object(
                Bucket=self._s3_bucket, Key=key)
            encoded = response['Body'].read()
            if encoded == None:
                return None
            if self.config.encoding == "json":
                return json.loads(encoded.decode('utf-8'))
            if self.config.encoding == "pickle":
                return pickle.loads(encoded)

    def set(self, key, value, ttl=600):
        if self.backend == "redis":
            if self.config.encoding == "json":
                encoded = json.dumps(value)
            elif self.config.encoding == "pickle":
                encoded = pickle.dumps(value)
            else:
                raise ValueError(
                    "redis config encoding must be one of the following:", VALID_ENCODINGS)
            self._redis_client.set(key, encoded, ex=ttl)

        if self.backend == "s3":
            if self.config.encoding == "json":
                encoded = json.dumps(value)
            elif self.config.encoding == "pickle":
                encoded = pickle.dumps(value)
            else:
                raise ValueError(
                    "s3 config encoding must be one of the following:", VALID_ENCODINGS)
            self._s3_client.put_object(
                Body=encoded, Bucket=self._s3_bucket, Key=key)

